/* Channel.h */

#ifndef __CHANNEL_H
#define __CHANNEL_H

#include	<string>
#include	<map>
#include	<iostream>

#include	<ctime>

#include	"ChannelUser.h"
#include	"xparameters.h"
#include	"Ban.h"
#include	"ELog.h"

using std::string ;
using std::map ;

namespace euworld
{

/// Forward declaration of class iClient.
class iClient ;

/**
 * This class represents a single network channel.
 * Channel users are also maintained in this structure.
 */
class Channel
{

protected:

	/**
	 * The type used to hold this channel's users.
	 */
	typedef map< unsigned int, ChannelUser* > userListType ;

	/// Make class xServer a friend of this class.
	friend class xServer ;

public:

	/**
	 * The type used to store this channel's current channel
	 * modes.
	 */
	typedef unsigned short int modeType ;

	/// Bit representing channel mode +t
	static const modeType	MODE_T  ;

	/// Bit representing channel mode +n
	static const modeType	MODE_N  ;

	/// Bit representing channel mode +s
	static const modeType	MODE_S  ;

	/// Bit representing channel mode +p
	static const modeType	MODE_P ;

	/// Bit representing channel mode +k
	static const modeType	MODE_K  ;

	/// Bit representing channel mode +l
	static const modeType	MODE_L  ;

	/// Bit representing channel mode +m
	static const modeType	MODE_M  ;

	/// Bit representing channel mode +i
	static const modeType	MODE_I ;

	/// Type used to store number of clients in channel
	typedef userListType::size_type size_type ;

	/**
	 * The type used for mutable iteration through this
	 * channel's user structure.
	 */
	typedef userListType::iterator userIterator ;

	/**
	 * The type used for immutable iteration through this
	 * channel's user structure.
	 */
	typedef userListType::const_iterator const_userIterator ;

	/**
	 * Construct a channel of the given name, constructed at the
	 * given creation time.
	 */
	Channel( const string& _name,
		const time_t& _creationTime ) ;

	/**
	 * Destroy this channel.
	 */
	virtual ~Channel() ;

	/**
	 * Test if a given channel mode is set.
	 */
	inline bool getMode( const modeType& whichMode ) const
		{ return (modes & whichMode) ; }

	/**
	 * Set the given channel mode.
	 */
	inline void setMode( const modeType& whichMode )
		{ modes |= whichMode ; }

	/**
	 * Remove the given channel mode.
	 */
	inline void removeMode( const modeType& whichMode )
		{ modes &= ~whichMode ; }

	/**
	 * Return true if the given mode is set for the the ChannelUser
	 * corresponding to the given iClient.
	 */
	inline bool getUserMode( const ChannelUser::modeType& whichMode,
		iClient* ) const ;

	/**
	 * Remove the given mode from the ChannelUser associated with the given
	 * iClient.
	 */
	inline bool removeUserMode( const ChannelUser::modeType& whichMode, iClient* ) ;

	/**
	 * Set the given mode for the ChannelUser associated with the given iClient.
	 */
	inline bool setUserMode( const ChannelUser::modeType& whichMode, iClient* ) ;

	/**
	 * Set a limit on the channel.  This method will set the
	 * channel mode and set the limit.
	 */
	inline void setLimit( const unsigned int& newLimit )
		{ limit = newLimit ; setMode( MODE_L ) ; }

	/**
	 * Set a key on the channel.  This method will set the
	 * channel mode and the key as well.
 	 */
	inline void setKey( const string& newKey )
		{ key = newKey ; setMode( MODE_K ) ; }

	/**
	 * Add a ban to this Channel's ban list.
	 */
	inline void setBan( const string& banMask ) ;

	/**
	 * Remove a ban from this channel's ban list.  This does a lexical
	 * comparison, not a wildcard match.
	 */
	inline void removeBan( const string& banMask ) ;

	/**
	 * Find a ban in the channel's ban list which lexically matches
	 * the given banMask.
	 */
	inline const Ban* findBan( const string& banMask ) const ;

	/**
	 * Find a ban in the channel's ban list which wildcard matches
	 * the given banMask.
	 */
	inline const Ban* matchBan( const string& banMask ) const ;

	/**
	 * Retrieve the current channel modes.
	 */
	inline const modeType& getModes() const
		{ return modes ; }

	inline const string getModeString() const ;

	/**
	 * Retrieve the name of this channel.
	 */
	inline const string& getName() const
		{ return name ; }

	/**
	 * Retrieve the creation time of this channel.
	 */
	inline const time_t& getCreationTime() const
		{ return creationTime ; }

	/**
	 * Retrieve this channel's key.  Note that the
	 * existence of a key does not mean that channel
	 * mode +k is set.
	 */
	inline const string& getKey() const
		{ return key ; }

	/**
	 * Retrieve this channel's limit.  Note that the
	 * existence of a limit here does not mean that
	 * channel mode +l is set.
	 */
	inline const unsigned int& getLimit() const
		{ return limit ; }

	/**
	 * Return true if no clients remain in this channel,
	 * false otherwise.
	 */
	inline bool empty() const
		{ return userList.empty() ; }

	/**
	 * Return the number of clients in this channel.
	 */
	inline size_type size() const
		{ return userList.size() ; }

	/**
	 * Retrieve a const iterator to the beginning of
	 * this channel's user structure.
	 */
	inline const_userIterator userList_begin() const
		{ return userList.begin() ; }

	/**
	 * Retrieve a const iterator to the end of this
	 * channel's user structure.
	 */
	inline const_userIterator userList_end() const
		{ return userList.end() ; }

	/**
	 * Add a ChannelUser to this channel's internal user
	 * structure.
	 */
	inline bool addUser( ChannelUser* newUser ) ;

	/**
	 * Remove a ChannelUser from this channel's internal
	 * user structure.  Note that the memory associated
	 * with the ChannelUser is NOT deallocated, it is
	 * returned to the caller.
	 */
	inline ChannelUser* removeUser( iClient* theClient ) ;

	/**
	 * Remove a ChannelUser from this channel's internal
	 * user structure.  Note that the memory associated
	 * with the ChannelUser is NOT deallocated, it is
	 * returned to the caller.
	 */
	inline ChannelUser* removeUser( const unsigned int& intYYXXX ) ;

	/**
	 * Return the ChannelUser associated with the given iClient,
	 * NULL if not found.
	 */
	inline ChannelUser* findUser( iClient* theClient ) const ;

	/**
	 * This method handles one or more channel mode changes.
	 * It is public here because it may be called from client space.
	 */
	virtual void	OnModeChange( const string yyxxx,
				const string& modes, const string& args ) ;

	/**
	 * Convenience operator for outputting Channel information
	 * to a C++ output stream.
	 */
	friend ELog& operator<<( ELog& out, const Channel& rhs )
	{
	out	<< "Name: " << rhs.name << endl
		<< "Creation time: " << rhs.creationTime ;
	return out ;
	}

protected:

	/**
	 * This method handles one or more channel mode changes.
	 * It is protected here because it is formatted as an xServer
	 * MSG handler method, and will be called by xServer.
	 */
	virtual void	OnModeChange( const xParameters& ) ;

	/**
	 * Set the creation time of this channel.  This is protected
	 * so that only class xServer may access it externally.
	 */
	inline virtual void	setCreationTime( const time_t& newCT )
		{ creationTime = newCT ; }

	/**
	 * The name of this channel.
	 */
	string		name ;

	/**
	 * The time at which this channel was created.
	 */
	time_t		creationTime ;
	
	/**
	 * This channel's current modes.
	 */
	modeType	modes ;

	/**
	 * The limit associated with this channel.
	 * Note that this variable may hold values
	 * even when channel mode +l is NOT set.
	 */
	unsigned int	limit ;

	/**
	 * The key associated with this channel.
	 * Note that this variable may hold values
	 * even when channel mode +k is NOT set.
	 */
	string		key ;

	/**
	 * The structure used to hold the ChannelUser
	 * instances.
	 */
	userListType	userList ;

} ;

} // namespace euworld

#endif // __CHANNEL_H
